from clickhouse_connect.datatypes import registry
from clickhouse_connect.driver.common import write_leb128
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import QueryResult, QueryContext
from clickhouse_connect.driver.types import ByteSource
from clickhouse_connect.driver.compression import get_compressor

_EMPTY_QUERY_CONTEXT = QueryContext()


class NativeTransform:
    # pylint: disable=too-many-locals
    @staticmethod
    def parse_response(source: ByteSource, context: QueryContext = _EMPTY_QUERY_CONTEXT) -> QueryResult:
        with context:
            names = []
            col_types = []
            use_none = context.use_none
            block_num = 0

            def get_block():
                nonlocal block_num
                result_block = []
                try:
                    num_cols = source.read_leb128()
                    num_rows = source.read_leb128()
                except (StopIteration, IndexError):
                    source.close()
                    return None
                for col_num in range(num_cols):
                    name = source.read_leb128_str()
                    type_name = source.read_leb128_str()
                    if block_num == 0:
                        names.append(name)
                        col_type = registry.get_from_name(type_name)
                        col_types.append(col_type)
                    else:
                        col_type = col_types[col_num]
                    context.start_column(name, col_type)
                    result_block.append(col_type.read_native_column(source, num_rows, use_none=use_none))
                block_num += 1
                return result_block

            first_block = get_block()
            if first_block is None:
                return QueryResult([])

            def gen():
                yield first_block
                while True:
                    next_block = get_block()
                    if next_block is None:
                        return
                    yield next_block

        return QueryResult(None, gen(), tuple(names), tuple(col_types), context.column_oriented, source)

    @staticmethod
    def build_insert(context: InsertContext):
        compressor = get_compressor(context.compression)

        with context:

            def chunk_gen():
                for x in context.next_block():
                    output = bytearray()
                    write_leb128(x.column_count, output)
                    write_leb128(x.row_count, output)
                    for col_name, col_type, data in zip(x.column_names, x.column_types, x.column_data):
                        write_leb128(len(col_name), output)
                        output += col_name.encode()
                        write_leb128(len(col_type.name), output)
                        output += col_type.name.encode()
                        context.start_column(col_name, col_type)
                        try:
                            col_type.write_native_column(data, output)
                        except Exception as ex:  # pylint: disable=broad-except
                            # This is hideous, but some low level serializations can fail while streaming
                            # the insert if the user has included bad data in the column.  We need to ensure that the
                            # insert fails (using garbage data) to avoid a partial insert, and use the context to
                            # propagate the correct exception to the user
                            context.insert_exception = ex
                            yield 'INTERNAL EXCEPTION WHILE SERIALIZING'.encode()
                            return
                    yield compressor.compress_block(output)
                footer = compressor.flush()
                if footer:
                    yield footer

        return chunk_gen()



import ipaddress
import logging
import re
import uuid
import pytz

from enum import Enum
from typing import Any, Tuple, Dict, Sequence, Optional, Union, Generator, Iterator
from datetime import date, datetime, tzinfo

from clickhouse_connect import common
from clickhouse_connect.driver.common import dict_copy
from clickhouse_connect.driver.types import Matrix, Closable
from clickhouse_connect.json_impl import any_to_json
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver.options import check_pandas, check_numpy, check_arrow
from clickhouse_connect.driver.context import BaseQueryContext

logger = logging.getLogger(__name__)
commands = 'CREATE|ALTER|SYSTEM|GRANT|REVOKE|CHECK|DETACH|DROP|DELETE|KILL|' + \
           'OPTIMIZE|SET|RENAME|TRUNCATE|USE'

limit_re = re.compile(r'\s+LIMIT($|\s)', re.IGNORECASE)
select_re = re.compile(r'(^|\s)SELECT\s', re.IGNORECASE)
insert_re = re.compile(r'(^|\s)INSERT\s*INTO', re.IGNORECASE)
command_re = re.compile(r'(^\s*)(' + commands + r')\s', re.IGNORECASE)
external_bind_re = re.compile(r'{.+:.+}')


# pylint: disable=too-many-instance-attributes
class QueryContext(BaseQueryContext):
    """
    Argument/parameter object for queries.  This context is used to set thread/query specific formats
    """

    # pylint: disable=duplicate-code,too-many-arguments
    def __init__(self,
                 query: str = '',
                 parameters: Optional[Dict[str, Any]] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                 encoding: Optional[str] = None,
                 server_tz: tzinfo = pytz.UTC,
                 use_none: Optional[bool] = None,
                 column_oriented: Optional[bool] = None):
        """
        Initializes various configuration settings for the query context

        :param query:  Query string with Python style format value replacements
        :param parameters: Optional dictionary of substitution values
        :param settings: Optional ClickHouse settings for the query
        :param query_formats: Optional dictionary of query formats with the key of a ClickHouse type name
          (with * wildcards) and a value of valid query formats for those types.
          The value 'encoding' can be sent to change the expected encoding for this query, with a value of
          the desired encoding such as `latin-1`
        :param column_formats: Optional dictionary of column specific formats.  The key is the column name,
          The value is either the format for the data column (such as 'string' for a UUID column) or a
          second level "format" dictionary of a ClickHouse type name and a value of query formats.  This
          secondary dictionary can be used for nested column types such as Tuples or Maps
        :param encoding: Optional string encoding for this query, such as 'latin-1'
        :param column_formats: Optional dictionary
        :param use_none:
        """
        super().__init__(settings, query_formats, column_formats, encoding)
        self.query = query
        self.parameters = parameters or {}
        self.server_tz = server_tz
        self.use_none = True if use_none is None else use_none
        self.column_oriented = False if column_oriented is None else column_oriented
        self._update_query()

    @property
    def is_select(self) -> bool:
        return select_re.search(self.uncommented_query) is not None

    @property
    def has_limit(self) -> bool:
        return limit_re.search(self.uncommented_query) is not None

    @property
    def is_insert(self) -> bool:
        return insert_re.search(self.uncommented_query) is not None

    @property
    def is_command(self) -> bool:
        return command_re.search(self.uncommented_query) is not None

    def set_parameters(self, parameters: Dict[str, Any]):
        self.parameters = parameters
        self._update_query()

    def set_parameter(self, key: str, value: Any):
        if not self.parameters:
            self.parameters = {}
        self.parameters[key] = value
        self._update_query()

    def updated_copy(self,
                     query: Optional[str] = None,
                     parameters: Optional[Dict[str, Any]] = None,
                     settings: Optional[Dict[str, Any]] = None,
                     query_formats: Optional[Dict[str, str]] = None,
                     column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                     encoding: Optional[str] = None,
                     server_tz: Optional[tzinfo] = None,
                     use_none: Optional[bool] = None,
                     column_oriented: Optional[bool] = None) -> 'QueryContext':
        """
        Creates Query context copy with parameters overridden/updated as appropriate.
        """
        return QueryContext(query or self.query,
                            dict_copy(self.parameters, parameters),
                            dict_copy(self.settings, settings),
                            dict_copy(self.query_formats, query_formats),
                            dict_copy(self.column_formats, column_formats),
                            encoding if encoding else self.encoding,
                            server_tz if server_tz else self.server_tz,
                            self.use_none if use_none is None else use_none,
                            self.column_oriented if column_oriented is None else column_oriented)

    def _update_query(self):
        self.final_query, self.bind_params = bind_query(self.query, self.parameters, self.server_tz)
        self.uncommented_query = remove_sql_comments(self.final_query)


class QueryResult:
    """
    Wrapper class for query return values and metadata
    """

    # pylint: disable=too-many-arguments
    def __init__(self,
                 result_set: Matrix = None,
                 block_gen: Generator[Matrix, None, None] = None,
                 column_names: Tuple[str, ...] = (),
                 column_types: Tuple[ClickHouseType, ...] = (),
                 column_oriented: bool = False,
                 source: Closable = None,
                 query_id: str = None,
                 summary: Dict[str, Any] = None):
        self._result_rows = result_set
        self._result_columns = None
        self._block_gen = block_gen
        self._in_context = False
        self.column_names = column_names
        self.column_types = column_types
        self.column_oriented = column_oriented
        self.source = source
        self.query_id = query_id
        self.summary = {} if summary is None else summary

    @property
    def empty(self):
        return self.row_count == 0

    @property
    def result_set(self) -> Matrix:
        if self.column_oriented:
            return self.result_columns
        return self.result_rows

    @property
    def result_columns(self) -> Matrix:
        if self._result_columns is None:
            if self._result_rows is not None:
                raise ProgrammingError(
                    'result_columns referenced after result_rows.  Only one final format is supported'
                )
            result = [[] for _ in range(len(self.column_names))]
            for block in self._block_gen:
                for base, added in zip(result, block):
                    base.extend(added)
            self._result_columns = result
            self._block_gen = None
        return self._result_columns

    @property
    def result_rows(self) -> Matrix:
        if self._result_rows is None:
            if self._result_columns is not None:
                raise ProgrammingError(
                    'result_rows referenced after result_columns.  Only one final format is supported'
                )
            result = []
            for block in self._block_gen:
                result.extend(list(zip(*block)))
            self._result_rows = result
            self._block_gen = None
        return self._result_rows

    def stream_column_blocks(self) -> Iterator[Matrix]:
        if not self._in_context:
            logger.warning("Streaming results should be used in a 'with' context to ensure the stream is closed")
        if not self._block_gen:
            raise ProgrammingError('Stream closed')
        temp = self._block_gen
        self._block_gen = None
        return temp

    def stream_row_blocks(self):
        return (list(zip(*block)) for block in self.stream_column_blocks())

    def stream_rows(self) -> Iterator[Sequence]:
        for block in self.stream_column_blocks():
            for row in list(zip(*block)):
                yield row

    def named_results(self) -> Generator[dict, None, None]:
        for row in zip(*self.result_columns):
            yield dict(zip(self.column_names, row))

    @property
    def row_count(self) -> int:
        if self.column_oriented:
            return 0 if len(self.result_set) == 0 else len(self.result_set[0])
        return len(self.result_set)

    @property
    def first_item(self):
        if self.empty:
            return None
        if self.column_oriented:
            return {name: col[0] for name, col in zip(self.column_names, self.result_set)}
        return dict(zip(self.column_names, self.result_set[0]))

    @property
    def first_row(self):
        if self.empty:
            return None
        if self.column_oriented:
            return [col[0] for col in self.result_set]
        return self.result_set[0]

    def close(self):
        if self.source:
            self.source.close()
            self.source = None

    def __enter__(self):
        self._in_context = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        self._in_context = False


local_tz = datetime.now().astimezone().tzinfo
BS = '\\'
must_escape = (BS, '\'')


def quote_identifier(identifier: str):
    first_char = identifier[0]
    if first_char in ('`', '"') and identifier[-1] == first_char:
        # Identifier is already quoted, assume that it's valid
        return identifier
    return f'`{identifier}`'


def finalize_query(query: str, parameters: Optional[Union[Sequence, Dict[str, Any]]],
                   server_tz: Optional[tzinfo] = None) -> str:
    if not parameters:
        return query
    if hasattr(parameters, 'items'):
        return query % {k: format_query_value(v, server_tz) for k, v in parameters.items()}
    return query % tuple(format_query_value(v) for v in parameters)


def bind_query(query: str, parameters: Optional[Union[Sequence, Dict[str, Any]]],
               server_tz: Optional[tzinfo] = None) -> Tuple[str, Dict[str, str]]:
    if not parameters:
        return query, {}
    if external_bind_re.search(query) is None:
        return finalize_query(query, parameters, server_tz), {}
    return query, {f'param_{k}': format_bind_value(v, server_tz) for k, v in parameters.items()}


def format_str(value: str):
    return f"'{''.join(f'{BS}{c}' if c in must_escape else c for c in value)}'"


# pylint: disable=too-many-return-statements
def format_query_value(value: Any, server_tz: tzinfo = pytz.UTC):
    """
    Format Python values in a ClickHouse query
    :param value: Python object
    :param server_tz: Server timezone for adjusting datetime values
    :return: Literal string for python value
    """
    if value is None:
        return 'NULL'
    if isinstance(value, str):
        return format_str(value)
    if isinstance(value, datetime):
        if value.tzinfo is None and server_tz != local_tz:
            value = value.replace(tzinfo=server_tz)
        return f"'{value.strftime('%Y-%m-%d %H:%M:%S')}'"
    if isinstance(value, date):
        return f"'{value.isoformat()}'"
    if isinstance(value, list):
        return f"[{', '.join(format_query_value(x, server_tz) for x in value)}]"
    if isinstance(value, tuple):
        return f"({', '.join(format_query_value(x, server_tz) for x in value)})"
    if isinstance(value, dict):
        if common.get_setting('dict_parameter_format') == 'json':
            return format_str(any_to_json(value).decode())
        pairs = [format_query_value(k, server_tz) + ':' + format_query_value(v, server_tz)
                 for k, v in value.items()]
        return f"{{{', '.join(pairs)}}}"
    if isinstance(value, Enum):
        return format_query_value(value.value, server_tz)
    if isinstance(value, (uuid.UUID, ipaddress.IPv4Address, ipaddress.IPv6Address)):
        return f"'{value}'"
    return str(value)


def format_bind_value(value: Any, server_tz: tzinfo = pytz.UTC):
    """
    Format Python values in a ClickHouse query
    :param value: Python object
    :param server_tz: Server timezone for adjusting datetime values
    :return: Literal string for python value
    """
    if value is None:
        return 'NULL'
    if isinstance(value, datetime):
        if value.tzinfo is None and server_tz != local_tz:
            value = value.replace(tzinfo=server_tz)
        return value.strftime('%Y-%m-%d %H:%M:%S')
    if isinstance(value, date):
        return value.isoformat()
    if isinstance(value, list):
        return f"[{', '.join(format_bind_value(x, server_tz) for x in value)}]"
    if isinstance(value, tuple):
        return f"({', '.join(format_bind_value(x, server_tz) for x in value)})"
    if isinstance(value, dict):
        if common.get_setting('dict_parameter_format') == 'json':
            return any_to_json(value).decode()
        pairs = [format_bind_value(k, server_tz) + ':' + format_bind_value(v, server_tz)
                 for k, v in value.items()]
        return f"{{{', '.join(pairs)}}}"
    if isinstance(value, Enum):
        return format_bind_value(value.value, server_tz)
    return str(value)


comment_re = re.compile(r"(\".*?\"|\'.*?\')|(/\*.*?\*/|(--\s)[^\n]*$)", re.MULTILINE | re.DOTALL)


def remove_sql_comments(sql: str) -> str:
    """
    Remove SQL comments.  This is useful to determine the type of SQL query, such as SELECT or INSERT, but we
    don't fully trust it to correctly ignore weird quoted strings, and other edge cases, so we always pass the
    original SQL to ClickHouse (which uses a full-fledged AST/ token parser)
    :param sql:  SQL query
    :return: SQL Query without SQL comments
    """

    def replacer(match):
        # if the 2nd group (capturing comments) is not None, it means we have captured a
        # non-quoted, actual comment string, so return nothing to remove the comment
        if match.group(2):
            return ''
        # Otherwise we've actually captured a quoted string, so return it
        return match.group(1)

    return comment_re.sub(replacer, sql)


def np_result(result: QueryResult, use_none: bool = False, max_str_len: int = 0):
    """
    See doc string from client.query_np
    """
    np = check_numpy()
    if result.empty:
        return np.empty(0)
    if not result.column_oriented:
        raise ProgrammingError('Numpy arrays should only be constructed from column oriented query results')
    np_types = [col_type.np_type(max_str_len) for col_type in result.column_types]
    first_type = np.dtype(np_types[0])
    if first_type != np.object_ and all(np.dtype(np_type) == first_type for np_type in np_types):
        # Optimize the underlying "matrix" array without any additional processing
        return np.array(result.result_set, first_type).transpose()
    columns = []
    has_objects = False
    for column, col_type, np_type in zip(result.result_set, result.column_types, np_types):
        if np_type == 'O':
            columns.append(column)
            has_objects = True
        elif use_none and col_type.nullable:
            new_col = []
            item_array = np.empty(1, dtype=np_type)
            for x in column:
                if x is None:
                    new_col.append(None)
                    has_objects = True
                else:
                    item_array[0] = x
                    new_col.append(item_array[0])
            columns.append(new_col)
        elif 'date' in np_type:
            columns.append(np.array(column, dtype=np_type))
        else:
            columns.append(column)
    if has_objects:
        np_types = [np.object_] * len(result.column_names)
    dtypes = np.dtype(list(zip(result.column_names, np_types)))
    return np.rec.fromarrays(columns, dtypes)


def pandas_result(result: QueryResult):
    """
    Convert QueryResult to a pandas dataframe
    :param result: QueryResult from driver
    :return: Two dimensional pandas dataframe from result
    """
    pd = check_pandas()
    np = check_numpy()
    if not result.column_oriented:
        raise ProgrammingError('Pandas dataframes should only be constructed from column oriented query results')
    raw = {}
    for name, col_type, column in zip(result.column_names, result.column_types, result.result_set):
        np_type = col_type.np_type()
        if 'datetime' in np_type:
            column = pd.to_datetime(np.array(column, dtype=np_type))
        raw[name] = column
    return pd.DataFrame(raw)


def to_arrow(content: bytes):
    pyarrow = check_arrow()
    reader = pyarrow.ipc.RecordBatchFileReader(content)
    return reader.read_all()


def arrow_buffer(table) -> Tuple[Sequence[str], bytes]:
    pyarrow = check_arrow()
    sink = pyarrow.BufferOutputStream()
    with pyarrow.RecordBatchFileWriter(sink, table.schema) as writer:
        writer.write(table)
    return table.schema.names, sink.getvalue()



from typing import NamedTuple, Sequence

from clickhouse_connect.datatypes.base import ClickHouseType


class TableColumnDef(NamedTuple):
    """
    Simplified ClickHouse Table Column definition for DDL
    """
    name: str
    ch_type: ClickHouseType
    expr_type: str = None
    expr: str = None

    @property
    def col_expr(self):
        expr = f'{self.name} {self.ch_type.name}'
        if self.expr_type:
            expr += f' {self.expr_type} {self.expr}'
        return expr


def create_table(table_name: str, columns: Sequence[TableColumnDef], engine: str, engine_params: dict):
    stmt = f"CREATE TABLE {table_name} ({', '.join(col.col_expr for col in columns)}) ENGINE {engine} "
    if engine_params:
        for key, value in engine_params.items():
            stmt += f' {key} {value}'
    return stmt



import logging
from inspect import signature
from typing import Optional, Union, Dict, Any
from urllib.parse import urlparse, parse_qs

from clickhouse_connect.driver.buffer import ResponseBuffer

from clickhouse_connect.driver.common import dict_copy

from clickhouse_connect.driver.exceptions import ProgrammingError
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.httpclient import HttpClient

logger = logging.getLogger(__name__)
Client.BuffCls = ResponseBuffer
try:
    from clickhouse_connect.driverc.buffer import ResponseBuffer as CResponseBuffer
    Client.BuffCls = CResponseBuffer
except ImportError:
    CResponseBuffer = None
    logger.warning('Unable to connect optimized C driver functions, falling back to pure Python', exc_info=True)


# pylint: disable=too-many-arguments,too-many-locals,too-many-branches
def create_client(host: str = None,
                  username: str = None,
                  password: str = '',
                  database: str = '__default__',
                  interface: Optional[str] = None,
                  port: int = 0,
                  secure: Union[bool, str] = False,
                  dsn: Optional[str] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  generic_args: Optional[Dict[str, Any]] = None,
                  **kwargs) -> Client:
    if dsn:
        parsed = urlparse(dsn)
        username = username or parsed.username
        password = password or parsed.password
        host = host or parsed.hostname
        port = port or parsed.port
        if parsed.path and not database:
            database = parsed.path[1:].split('/')[0]
        database = database or parsed.path
        kwargs.update(dict(parse_qs(parsed.query)))
    use_tls = str(secure).lower() == 'true' or interface == 'https' or (not interface and port in (443, 8443))
    if not host:
        host = 'localhost'
    if not interface:
        interface = 'https' if use_tls else 'http'
    port = port or default_port(interface, use_tls)
    if username is None and 'user' in kwargs:
        username = kwargs.pop('user')
    if password and username is None:
        username = 'default'
    if 'compression' in kwargs and 'compress' not in kwargs:
        kwargs['compress'] = kwargs.pop('compression')
    settings = settings or {}
    if interface.startswith('http'):
        if generic_args:
            client_params = signature(HttpClient).parameters
            for name, value in generic_args.items():
                if name in client_params:
                    kwargs[name] = value
                else:
                    if name.startswith('ch_'):
                        name = name[3:]
                    settings[name] = value
        return HttpClient(interface, host, port, username, password, database, settings=settings, **kwargs)
    raise ProgrammingError(f'Unrecognized client type {interface}')


def default_port(interface: str, secure: bool):
    if interface.startswith('http'):
        return 8443 if secure else 8123
    raise ValueError('Unrecognized ClickHouse interface')



import pytz

from datetime import date, timedelta, datetime
from typing import Union, Sequence, MutableSequence

from clickhouse_connect.datatypes.base import TypeDef, ArrayType
from clickhouse_connect.driver.common import  write_array, np_date_types
from clickhouse_connect.driver.types import ByteSource

epoch_start_date = date(1970, 1, 1)
epoch_start_datetime = datetime(1970, 1, 1)


class Date(ArrayType):
    _array_type = 'H'
    _np_type = 'datetime64[D]'
    nano_divisor = 86400 * 1000000000
    valid_formats = 'native', 'int'
    python_null = epoch_start_date
    python_type = date

    def _read_native_binary(self, source: ByteSource, num_rows: int):
        column = source.read_array(self._array_type, num_rows)
        if self.read_format() == 'int':
            return column
        return [epoch_start_date + timedelta(days) for days in column]

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        first = self._first_value(column)
        if isinstance(first, int) or self.write_format() == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        else:
            if isinstance(first, datetime):
                esd = epoch_start_datetime
            else:
                esd = epoch_start_date
            if self.nullable:
                column = [0 if x is None else (x - esd).days for x in column]
            else:
                column = [(x - esd).days for x in column]
        write_array(self._array_type, column, dest)


class Date32(Date):
    _array_type = 'i'


from_ts_naive = datetime.utcfromtimestamp
from_ts_tz = datetime.fromtimestamp


# pylint: disable=abstract-method
class DateTime(ArrayType):
    _array_type = 'I'
    _np_type = 'datetime64[s]'
    valid_formats = 'native', 'int'
    python_null = from_ts_naive(0)
    python_type = datetime
    nano_divisor = 1000000000

    def _read_native_binary(self, source: ByteSource, num_rows: int):
        column = source.read_array(self._array_type, num_rows)
        if self.read_format() == 'int':
            return column
        fts = from_ts_naive
        return [fts(ts) for ts in column]

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        first = self._first_value(column)
        if isinstance(first, int) or self.write_format() == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        else:
            if self.nullable:
                column = [int(x.timestamp()) if x else 0 for x in column]
            else:
                column = [int(x.timestamp()) for x in column]
        write_array(self._array_type, column, dest)


class DateTime64(ArrayType):
    __slots__ = 'scale', 'prec', 'tzinfo'
    _array_type = 'Q'
    valid_formats = 'native', 'int'
    python_null = epoch_start_date
    python_type = datetime

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str
        self.scale = type_def.values[0]
        self.prec = 10 ** self.scale
        if len(type_def.values) > 1:
            self.tzinfo = pytz.timezone(type_def.values[1][1:-1])
            self._read_native_binary = self._read_native_tz
        else:
            self._read_native_binary = self._read_native_naive
            self.tzinfo = None

    def np_type(self, _str_len: int = 0):
        opt = np_date_types.get(self.scale)
        return f'datetime64{opt}' if opt else 'O'

    @property
    def nano_divisor(self):
        return 1000000000 // self.prec

    def _read_native_tz(self, source: ByteSource, num_rows: int):
        column = source.read_array(self._array_type, num_rows)
        if self.read_format() == 'int':
            return column
        new_col = []
        app = new_col.append
        dt_from = datetime.fromtimestamp
        prec = self.prec
        tz_info = self.tzinfo
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds, tz_info)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col

    def _read_native_naive(self, source: ByteSource, num_rows: int):
        column = source.read_array(self._array_type, num_rows)
        if self.read_format() == 'int':
            return column
        new_col = []
        app = new_col.append
        dt_from = datetime.utcfromtimestamp
        prec = self.prec
        for ticks in column:
            seconds = ticks // prec
            dt_sec = dt_from(seconds)
            app(dt_sec.replace(microsecond=((ticks - seconds * prec) * 1000000) // prec))
        return new_col

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        first = self._first_value(column)
        if isinstance(first, int) or self.write_format() == 'int':
            if self.nullable:
                column = [x if x else 0 for x in column]
        else:
            prec = self.prec
            if self.nullable:
                column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 if x else 0 for x in column]
            else:
                column = [((int(x.timestamp()) * 1000000 + x.microsecond) * prec) // 1000000 for x in column]
        write_array(self._array_type, column, dest)



import http
import logging
import sys
import socket

import certifi
import lz4.frame
import zstandard
from urllib3.poolmanager import PoolManager
from urllib3.response import HTTPResponse


# Increase this number just to be safe when ClickHouse is returning progress headers
http._MAXHEADERS = 10000  # pylint: disable=protected-access

DEFAULT_KEEP_INTERVAL = 30
DEFAULT_KEEP_COUNT = 3
DEFAULT_KEEP_IDLE = 30

SOCKET_TCP = socket.IPPROTO_TCP

core_socket_options = [
    (socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
    (SOCKET_TCP, socket.TCP_NODELAY, 1),
    (socket.SOL_SOCKET, socket.SO_SNDBUF, 1024 * 256),
    (socket.SOL_SOCKET, socket.SO_SNDBUF, 1024 * 256)
]

logging.getLogger('urllib3').setLevel(logging.WARNING)


# pylint: disable=no-member
def get_pool_manager(keep_interval: int = DEFAULT_KEEP_INTERVAL,
                     keep_count: int = DEFAULT_KEEP_COUNT,
                     keep_idle: int = DEFAULT_KEEP_IDLE,
                     ca_cert: str = None,
                     verify: bool = True,
                     client_cert: str = None,
                     client_cert_key: str = None,
                     **options) -> PoolManager:
    socket_options = core_socket_options.copy()
    if getattr(socket, 'TCP_KEEPINTVL', None) is not None:
        socket_options.append((SOCKET_TCP, socket.TCP_KEEPINTVL, keep_interval))
    if getattr(socket, 'TCP_KEEPCNT', None) is not None:
        socket_options.append((SOCKET_TCP, socket.TCP_KEEPCNT, keep_count))
    if getattr(socket, 'TCP_KEEPIDLE', None) is not None:
        socket_options.append((SOCKET_TCP, socket.TCP_KEEPIDLE, keep_idle))
    if sys.platform == 'darwin':
        socket_options.append((SOCKET_TCP, getattr(socket, 'TCP_KEEPALIVE', 0x10), keep_interval))
    if ca_cert == 'certifi':
        ca_cert = certifi.where()
    options['cert_reqs'] = 'CERT_REQUIRED' if verify else 'CERT_NONE'
    if ca_cert:
        options['ca_certs'] = ca_cert
    if client_cert:
        options['cert_file'] = client_cert
    if client_cert_key:
        options['key_file'] = client_cert_key
    return PoolManager(block=False, socket_options=socket_options, **options)


def get_response_data(response: HTTPResponse) -> bytes:
    encoding = response.headers.get('content-encoding', None)
    if encoding == 'zstd':
        try:
            zstd_decom = zstandard.ZstdDecompressor()
            return zstd_decom.stream_reader(response.data).read()
        except zstandard.ZstdError:
            pass
    if encoding == 'lz4':
        lz4_decom = lz4.frame.LZ4FrameDecompressor()
        return lz4_decom.decompress(response.data, len(response.data))
    return response.data


default_pool_manager = get_pool_manager()


class ResponseSource:
    def __init__(self, response: HTTPResponse, chunk_size: int = 1024 * 1024):
        self.response = response
        compression = response.headers.get('content-encoding')
        if compression == 'zstd':
            zstd_decom = zstandard.ZstdDecompressor()
            reader = zstd_decom.stream_reader(self, read_across_frames=False)

            def decompress():
                while True:
                    chunk = reader.read()
                    if not chunk:
                        break
                    yield chunk

            self.gen = decompress()
        elif compression == 'lz4':
            lz4_decom = lz4.frame.LZ4FrameDecompressor()

            def decompress():
                while lz4_decom.needs_input:
                    data = self.response.read(chunk_size)
                    if lz4_decom.unused_data:
                        data = lz4_decom.unused_data + data
                    if not data:
                        return
                    chunk = lz4_decom.decompress(data)
                    if chunk:
                        yield chunk

            self.gen = decompress()
        else:
            self.gen = response.stream(decode_content=True)

    def read(self, amt: int) -> bytes:
        return self.response.read(amt)

    def close(self):
        self.response.drain_conn()
        self.response.close()



import array
import logging

from abc import ABC
from math import log
from typing import NamedTuple, Dict, Type, Any, Sequence, MutableSequence, Optional, Union, Iterable

from clickhouse_connect.driver.common import array_type, int_size, write_array, write_uint64, low_card_version
from clickhouse_connect.driver.exceptions import NotSupportedError
from clickhouse_connect.driver.threads import query_settings
from clickhouse_connect.driver.types import ByteSource

logger = logging.getLogger(__name__)
ch_read_formats = {}
ch_write_formats = {}


class TypeDef(NamedTuple):
    """
    Immutable tuple that contains all additional information needed to construct a particular ClickHouseType
    """
    wrappers: tuple = ()
    keys: tuple = ()
    values: tuple = ()

    @property
    def arg_str(self):
        return f"({', '.join(str(v) for v in self.values)})" if self.values else ''


class ClickHouseType(ABC):
    """
    Base class for all ClickHouseType objects.
    """
    __slots__ = 'nullable', 'low_card', 'wrappers', 'type_def', '__dict__'
    _name_suffix = ''
    _encoding = 'utf8'
    _np_type = 'O'  # Default to Numpy Object type
    nano_divisor = 0  # Only relevant for date like objects
    byte_size = 0
    valid_formats = 'native'

    python_null = 0
    python_type = None
    base_type = None

    def __init_subclass__(cls, registered: bool = True):
        if registered:
            cls.base_type = cls.__name__
            type_map[cls.base_type] = cls

    @classmethod
    def build(cls: Type['ClickHouseType'], type_def: TypeDef):
        return cls(type_def)

    @classmethod
    def _active_format(cls, fmt_map: Dict[Type['ClickHouseType'], str]):
        overrides = getattr(query_settings, 'column_overrides', None)
        if overrides and cls in overrides:
            return overrides[cls]
        overrides = getattr(query_settings, 'query_overrides', None)
        if overrides and cls in overrides:
            return overrides[cls]
        return fmt_map.get(cls, 'native')

    @classmethod
    def read_format(cls):
        return cls._active_format(ch_read_formats)

    @classmethod
    def write_format(cls):
        return cls._active_format(ch_write_formats)

    def __init__(self, type_def: TypeDef):
        """
        Base class constructor that sets Nullable and LowCardinality wrappers
        :param type_def:  ClickHouseType base configuration parameters
        """
        self.type_def = type_def
        self.wrappers = type_def.wrappers
        self.low_card = 'LowCardinality' in self.wrappers
        self.nullable = 'Nullable' in self.wrappers

    def __eq__(self, other):
        return other.__class__ == self.__class__ and self.type_def == other.type_def

    def __hash__(self):
        return hash((self.type_def, self.__class__))

    @property
    def name(self):
        name = f'{self.base_type}{self._name_suffix}'
        for wrapper in reversed(self.wrappers):
            name = f'{wrapper}({name})'
        return name

    @property
    def encoding(self):
        query_encoding = getattr(query_settings, 'query_encoding', None)
        return query_encoding or self._encoding

    def np_type(self, _str_len: int = 0):
        return self._np_type

    def write_native_prefix(self, dest: MutableSequence):
        """
        Prefix is primarily used is for the LowCardinality version (but see the JSON data type).  Because of the
        way the ClickHouse C++ code is written, this must be done before any data is written even if the
        LowCardinality column is within a container.  The only recognized low cardinality version is 1
        :param dest: The native protocol binary write buffer
        """
        if self.low_card:
            write_uint64(low_card_version, dest)

    def read_native_prefix(self, source: ByteSource):
        """
        Read the low cardinality version.  Like the write method, this has to happen immediately for container classes
        :param source: The native protocol binary read buffer
        :return: updated read pointer
        """
        if self.low_card:
            v = source.read_uint64()
            if v != low_card_version:
                logger.warning('Unexpected low cardinality version %d reading type %s', v, self.name)

    def read_native_column(self, source: ByteSource, num_rows: int, **kwargs) -> Sequence:
        """
        Wrapping read method for all ClickHouseType data types.  Only overridden for container classes so that
         the LowCardinality version is read for the contained types
        :param source: Native protocol binary read buffer
        :param num_rows: Number of rows expected in the column
        :param kwargs: Pass any extra keyword arguments to the main read_native_data function
        :return: The decoded column data as a sequence and the updated location pointer
        """
        self.read_native_prefix(source)
        return self.read_native_data(source, num_rows, **kwargs)

    def read_native_data(self, source: ByteSource, num_rows: int, use_none=True) -> Sequence:
        """
        Public read method for all ClickHouseType data type columns
        :param source: Native protocol binary read buffer
        :param num_rows: Number of rows expected in the column
        :param use_none: Use the Python None type for ClickHouse nulls.  Otherwise, use the empty or zero type.
         Allows support for pandas data frames that do not support None
        :return: The decoded column plus the updated location pointer
        """
        if self.low_card:
            return self._read_native_low_card(source, num_rows, use_none)
        if self.nullable:
            null_map = source.read_bytes(num_rows)
            column = self._read_native_binary(source, num_rows)
            if use_none:
                if isinstance(column, (tuple, array.array)):
                    return [None if null_map[ix] else column[ix] for ix in range(num_rows)]
                for ix in range(num_rows):
                    if null_map[ix]:
                        column[ix] = None
            return column
        return self._read_native_binary(source, num_rows)

    # These two methods are really abstract, but they aren't implemented for container classes which
    # delegate binary operations to their elements

    # pylint: disable=no-self-use
    def _read_native_binary(self, _source: ByteSource, _num_rows: int) -> Union[Sequence, MutableSequence]:
        """
        Lowest level read method for ClickHouseType native data columns
        :param _source: Native protocol binary read buffer
        :param _num_rows: Expected number of rows in the column
        :return: Decoded column plus updated read buffer
        """
        return [], 0

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        """
        Lowest level write method for ClickHouseType data columns
        :param column: Python data column
        :param dest: Native protocol write buffer
        """

    def write_native_column(self, column: Sequence, dest: MutableSequence):
        """
        Wrapping write method for ClickHouseTypes.  Only overridden for container types that so that
        the write_native_prefix is done at the right time for contained types
        :param column: Column/sequence of Python values to write
        :param dest: Native binary write buffer
        """
        self.write_native_prefix(dest)
        self.write_native_data(column, dest)

    def write_native_data(self, column: Sequence, dest: MutableSequence):
        """
        Public native write method for ClickHouseTypes.  Delegates the actual write to either the LowCardinality
        write method or the _write_native_binary method of the type
        :param column: Sequence of Python data
        :param dest: Native binary write buffer
        """
        if self.low_card:
            self._write_native_low_card(column, dest)
        else:
            if self.nullable:
                dest += bytes([1 if x is None else 0 for x in column])
            self._write_native_binary(column, dest)

    def _read_native_low_card(self, source: ByteSource, num_rows: int, use_none=True):
        if num_rows == 0:
            return tuple()
        key_data = source.read_uint64()
        index_sz = 2 ** (key_data & 0xff)
        key_cnt = source.read_uint64()
        keys= self._read_native_binary(source, key_cnt)
        if self.nullable:
            try:
                keys[0] = None if use_none else self.python_null
            except TypeError:
                keys = (None if use_none else self.python_null,) + tuple(keys[1:])
        index_cnt = source.read_uint64()
        assert index_cnt == num_rows
        index = source.read_array(array_type(index_sz, False), num_rows)
        return tuple(keys[ix] for ix in index)

    def _write_native_low_card(self, column: Iterable, dest: MutableSequence):
        if not column:
            return
        index = []
        keys = []
        rev_map = {}
        rmg = rev_map.get
        if self.nullable:
            keys.append(None)
            key = 1
            for x in column:
                if x is None:
                    index.append(0)
                else:
                    ix = rmg(x)
                    if ix is None:
                        index.append(key)
                        keys.append(x)
                        rev_map[x] = key
                        key += 1
                    else:
                        index.append(ix)
        else:
            key = 0
            for x in column:
                ix = rmg(x)
                if ix is None:
                    index.append(key)
                    keys.append(x)
                    rev_map[x] = key
                    key += 1
                else:
                    index.append(ix)
        ix_type = int(log(len(keys), 2)) >> 3  # power of two bytes needed to store the total number of keys
        write_uint64((1 << 9) | (1 << 10) | ix_type, dest)  # Index type plus new dictionary (9) and additional keys(10)
        write_uint64(len(keys), dest)
        self._write_native_binary(keys, dest)
        write_uint64(len(index), dest)
        write_array(array_type(1 << ix_type, False), index, dest)

    def _first_value(self, column: Sequence) -> Optional[Any]:
        if self.nullable:
            return next((x for x in column if x is not None), None)
        if len(column):
            return column[0]
        return None


EMPTY_TYPE_DEF = TypeDef()
NULLABLE_TYPE_DEF = TypeDef(wrappers=('Nullable',))
LC_TYPE_DEF = TypeDef(wrappers=('LowCardinality',))
type_map: Dict[str, Type[ClickHouseType]] = {}


class ArrayType(ClickHouseType, ABC, registered=False):
    """
    ClickHouse type that utilizes Python array.array for fast reads and writes of binary data.
    array.array can only be used for ClickHouse types that can be translated into UInt64 or small integers,
    or Float32/64
    """
    _signed = True
    _array_type = None
    _struct_type = None
    valid_formats = 'string', 'native'
    python_type = int

    def __init_subclass__(cls, registered: bool = True):
        super().__init_subclass__(registered)
        if cls._array_type in ('i', 'I') and int_size == 2:
            cls._array_type = 'L' if cls._array_type.isupper() else 'l'
        if isinstance(cls._array_type, str) and cls._array_type:
            cls._struct_type = '<' + cls._array_type
            cls.byte_size = array.array(cls._array_type).itemsize

    def _read_native_binary(self, source: ByteSource, num_rows: int):
        column = source.read_array(self._array_type, num_rows)
        if self.read_format() == 'string':
            column = [str(x) for x in column]
        return column

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        if len(column) and self.nullable:
            first = column[0]
            try:
                column[0] = None
                for ix, x in enumerate(column):
                    if not x:
                        column[ix] = 0
                column[0] = first or 0
            except TypeError:
                column = [0 if x is None else x for x in column]
        write_array(self._array_type, column, dest)


class UnsupportedType(ClickHouseType, ABC, registered=False):
    """
    Base class for ClickHouse types that can't be serialized/deserialized into Python types.
    Mostly useful just for DDL statements
    """
    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self._name_suffix = type_def.arg_str

    def _read_native_binary(self, source: Sequence, num_rows: int):
        raise NotSupportedError(f'{self.name} deserialization not supported')

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        raise NotSupportedError(f'{self.name} serialization  not supported')



import json
import logging
import platform
import re
import uuid
from base64 import b64encode
from http.client import RemoteDisconnected
from typing import Optional, Dict, Any, Sequence, Union, List, Callable, Generator, BinaryIO
from urllib.parse import urlencode

from urllib3 import Timeout
from urllib3.exceptions import HTTPError
from urllib3.poolmanager import PoolManager
from urllib3.response import HTTPResponse

from clickhouse_connect import common
from clickhouse_connect.datatypes import registry
from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.driver.client import Client
from clickhouse_connect.driver.common import dict_copy, coerce_bool, coerce_int
from clickhouse_connect.driver.compression import available_compression
from clickhouse_connect.driver.exceptions import DatabaseError, OperationalError, ProgrammingError
from clickhouse_connect.driver.httputil import ResponseSource, get_pool_manager, get_response_data, default_pool_manager
from clickhouse_connect.driver.insert import InsertContext
from clickhouse_connect.driver.query import QueryResult, QueryContext, quote_identifier, bind_query
from clickhouse_connect.driver.transform import NativeTransform

logger = logging.getLogger(__name__)
columns_only_re = re.compile(r'LIMIT 0\s*$', re.IGNORECASE)


# pylint: disable=too-many-instance-attributes
class HttpClient(Client):
    params = {}
    valid_transport_settings = {'database', 'buffer_size', 'session_id', 'compress', 'decompress',
                                'session_timeout', 'session_check', 'query_id', 'quota_key', 'wait_end_of_query',
                                }
    optional_transport_settings = {'send_progress_in_http_headers', 'http_headers_progress_interval_ms',
                                   'enable_http_compression'}

    # pylint: disable=too-many-arguments,too-many-locals,too-many-branches,too-many-statements
    def __init__(self,
                 interface: str,
                 host: str,
                 port: int,
                 username: str,
                 password: str,
                 database: str,
                 compress: Union[bool, str] = True,
                 query_limit: int = 5000,
                 query_retries: int = 2,
                 connect_timeout: int = 10,
                 send_receive_timeout: int = 300,
                 client_name: Optional[str] = None,
                 send_progress: bool = True,
                 verify: bool = True,
                 ca_cert: Optional[str] = None,
                 client_cert: Optional[str] = None,
                 client_cert_key: Optional[str] = None,
                 session_id: Optional[str] = None,
                 settings: Optional[Dict[str, Any]] = None,
                 pool_mgr: Optional[PoolManager] = None):
        """
        Create an HTTP ClickHouse Connect client
        :param interface: http or https
        :param host: hostname
        :param port: host port
        :param username: ClickHouse user
        :param password: ClickHouse password
        :param database: Default database for the connection
        :param compress: Accept compressed HTTP type from server (brotli or gzip)
        :param query_limit: Default LIMIT on returned rows
        :param connect_timeout:  Timeout in seconds for the http connection
        :param send_receive_timeout: Read timeout in seconds for http connection
        :param client_name: Http user agent header value
        :param send_progress: Ask ClickHouse to send progress headers.  Used for summary and keep alive
        :param verify: Verify the server certificate in secure/https mode
        :param ca_cert: If verify is True, the file path to Certificate Authority root to validate ClickHouse server
         certificate, in .pem format.  Ignored if verify is False.  This is not necessary if the ClickHouse server
         certificate is trusted by the operating system.  To trust the maintained list of "global" public root
         certificates maintained by the Python 'certifi' package, set ca_cert to 'certifi'
        :param client_cert: File path to a TLS Client certificate in .pem format.  This file should contain any
          applicable intermediate certificates
        :param client_cert_key: File path to the private key for the Client Certificate.  Required if the private key
          is not included the Client Certificate key file
        :param session_id ClickHouse session id.  If not specified and the common setting 'autogenerate_session_id'
          is True, the client will generate a UUID1 session id
        :param settings Optional dictionary of ClickHouse setting values (str/value) for every connection request
        :param pool_mgr Optional urllib3 PoolManager for this client.  Useful for creating separate connection
          pools for multiple client endpoints for applications with many clients
        """
        self.url = f'{interface}://{host}:{port}'
        self.headers = {}
        ch_settings = settings or {}
        self.http = pool_mgr
        if interface == 'https':
            if client_cert:
                if not username:
                    raise ProgrammingError('username parameter is required for Mutual TLS authentication')
                self.headers['X-ClickHouse-User'] = username
                self.headers['X-ClickHouse-SSL-Certificate-Auth'] = 'on'
            verify = coerce_bool(verify)
            if not self.http and (ca_cert or client_cert or not verify):
                self.http = get_pool_manager(ca_cert=ca_cert,
                                             client_cert=client_cert,
                                             verify=verify,
                                             client_cert_key=client_cert_key)
        if not self.http:
            self.http = default_pool_manager

        if not client_cert and username:
            self.headers['Authorization'] = 'Basic ' + b64encode(f'{username}:{password}'.encode()).decode()

        client_name = client_name.strip() + ' ' if client_name else ''

        self.headers['User-Agent'] = f'{client_name}clickhouse-connect/{common.version()} (py/{platform.version()})'
        self._read_format = self._write_format = 'Native'
        self._transform = NativeTransform()

        connect_timeout, send_receive_timeout = coerce_int(connect_timeout), coerce_int(send_receive_timeout)
        self.timeout = Timeout(connect=connect_timeout, read=send_receive_timeout)
        self.query_retries = coerce_int(query_retries)
        self.http_retries = 1

        if coerce_bool(send_progress):
            ch_settings['wait_end_of_query'] = '1'
            # We can't actually read the progress headers, but we enable them so ClickHouse sends data
            # to keep the connection alive when waiting for long-running queries that don't return data
            # Accordingly we make sure it's always less than the read timeout
            ch_settings['send_progress_in_http_headers'] = '1'
            progress_interval = min(120000, (send_receive_timeout - 5) * 1000)
            ch_settings['http_headers_progress_interval_ms'] = str(progress_interval)

        if session_id:
            ch_settings['session_id'] = session_id
        elif 'session_id' not in ch_settings and common.get_setting('autogenerate_session_id'):
            ch_settings['session_id'] = str(uuid.uuid1())

        if coerce_bool(compress):
            compression = ','.join(available_compression)
            self.write_compression = available_compression[0]
        elif compress and compress not in ('False', 'false', '0'):
            if compress not in available_compression:
                raise ProgrammingError(f'Unsupported compression method {compress}')
            compression = compress
            self.write_compression = compress
        else:
            compression = None

        super().__init__(database=database, query_limit=coerce_int(query_limit), uri=self.url)

        comp_setting = self.server_settings['enable_http_compression']
        # We only set the header for the query method so no need to modify headers or settings here
        if comp_setting and (comp_setting.value == '1' or comp_setting.readonly != 1):
            self.compression = compression
        self.params = self._validate_settings(ch_settings)

    def set_client_setting(self, key, value):
        str_value = self._validate_setting(key, value, common.get_setting('invalid_setting_action'))
        if str_value is not None:
            self.params[key] = str_value

    def get_client_setting(self, key) -> Optional[str]:
        values = self.params.get(key)
        return values[0] if values else None

    def _prep_query(self, context: QueryContext):
        final_query = super()._prep_query(context)
        if context.is_insert:
            return final_query
        return f'{final_query}\n FORMAT {self._write_format}'

    def _query_with_context(self, context: QueryContext) -> QueryResult:
        headers = {'Content-Type': 'text/plain; charset=utf-8'}
        params = {}
        if self.database:
            params['database'] = self.database
        params.update(context.bind_params)
        params.update(self._validate_settings(context.settings))
        if columns_only_re.search(context.uncommented_query):
            response = self._raw_request(f'{context.final_query}\n FORMAT JSON',
                                         params, headers, retries=self.query_retries)
            json_result = json.loads(response.data)
            # ClickHouse will respond with a JSON object of meta, data, and some other objects
            # We just grab the column names and column types from the metadata sub object
            names: List[str] = []
            types: List[ClickHouseType] = []
            for col in json_result['meta']:
                names.append(col['name'])
                types.append(registry.get_from_name(col['type']))
            return QueryResult([], None, tuple(names), tuple(types))

        if self.compression:
            headers['Accept-Encoding'] = self.compression
            params['enable_http_compression'] = '1'
        response = self._raw_request(self._prep_query(context), params, headers, stream=True,
                                     retries=self.query_retries)
        byte_source = Client.BuffCls(ResponseSource(response)) # pylint: disable=not-callable
        query_result = self._transform.parse_response(byte_source, context)
        if 'X-ClickHouse-Summary' in response.headers:
            try:
                summary = json.loads(response.headers['X-ClickHouse-Summary'])
                query_result.summary = summary
            except json.JSONDecodeError:
                pass
        query_result.query_id = response.headers.get('X-ClickHouse-Query-Id')
        return query_result

    def data_insert(self, context: InsertContext):
        """
        See BaseClient doc_string for this method
        """
        if context.empty:
            logger.debug('No data included in insert, skipping')
            return
        if context.compression is None:
            context.compression = self.write_compression
        block_gen = self._transform.build_insert(context)

        def error_handler(response: HTTPResponse):
            # If we actually had a local exception when building the insert, throw that instead
            if context.insert_exception:
                ex = context.insert_exception
                context.insert_exception = None
                raise ProgrammingError('Internal serialization error.  This usually indicates invalid data types ' +
                                       'in an inserted row or column') from ex
            self._error_handler(response)

        self.raw_insert(context.table,
                        context.column_names,
                        block_gen,
                        context.settings,
                        self._write_format,
                        context.compression,
                        error_handler)
        context.data = None

    def raw_insert(self, table: str,
                   column_names: Optional[Sequence[str]] = None,
                   insert_block: Union[str, bytes, Generator[bytes, None, None], BinaryIO] = None,
                   settings: Optional[Dict] = None,
                   fmt: Optional[str] = None,
                   compression: Optional[str] = None,
                   status_handler: Optional[Callable] = None):
        """
        See BaseClient doc_string for this method
        """
        write_format = fmt if fmt else self._write_format
        headers = {'Content-Type': 'application/octet-stream'}
        if compression:
            headers['Content-Encoding'] = compression
        cols = f" ({', '.join([quote_identifier(x) for x in column_names])})" if column_names is not None else ''
        params = {'query': f'INSERT INTO {table}{cols} FORMAT {write_format}'}
        if self.database:
            params['database'] =  self.database
        params.update(self._validate_settings(settings or {}))
        response = self._raw_request(insert_block, params, headers, error_handler=status_handler)
        logger.debug('Insert response code: %d, content: %s', response.status, response.data)

    def command(self,
                cmd,
                parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                data: Union[str, bytes] = None,
                settings: Optional[Dict] = None,
                use_database: int = True) -> Union[str, int, Sequence[str]]:
        """
        See BaseClient doc_string for this method
        """
        cmd, params = bind_query(cmd, parameters, self.server_tz)
        headers = {}
        payload = None
        if isinstance(data, str):
            headers['Content-Type'] = 'text/plain; charset=utf-8'
            payload = data.encode()
        elif isinstance(data, bytes):
            headers['Content-Type'] = 'application/octet-stream'
            payload = data
        if payload is None:
            if not cmd:
                raise ProgrammingError('Command sent without query or recognized data') from None
            payload = cmd
        elif cmd:
            params['query'] = cmd
        if use_database and self.database:
            params['database'] = self.database
        params.update(self._validate_settings(settings or {}))
        method = 'POST' if payload else 'GET'
        response = self._raw_request(payload, params, headers, method)
        result = response.data.decode()[:-1].split('\t')
        if len(result) == 1:
            try:
                return int(result[0])
            except ValueError:
                return result[0]
        return result

    def _error_handler(self, response: HTTPResponse, retried: bool = False) -> None:
        err_str = f'HTTPDriver for {self.url} returned response code {response.status})'
        err_content = get_response_data(response)
        if err_content:
            err_msg = err_content.decode(errors='backslashreplace')
            logger.error(err_msg)
            err_str = f':{err_str}\n {err_msg[0:240]}'
        raise OperationalError(err_str) if retried else DatabaseError(err_str) from None

    def _raw_request(self,
                     data,
                     params: Dict[str, str],
                     headers: Optional[Dict[str, Any]] = None,
                     method: str = 'POST',
                     retries: int = 0,
                     stream: bool = False,
                     error_handler: Callable = None) -> HTTPResponse:
        if isinstance(data, str):
            data = data.encode()
        headers = dict_copy(self.headers, headers)
        url = f'{self.url}?{urlencode(dict_copy(self.params, params))}'
        attempts = 0
        while True:
            attempts += 1
            try:
                response: HTTPResponse = self.http.request(method, url,
                                                           headers=headers,
                                                           timeout=self.timeout,
                                                           body=data,
                                                           retries=self.http_retries,
                                                           preload_content=not stream)
            except HTTPError as ex:
                rex_context = ex.__context__
                if rex_context and isinstance(rex_context.__context__, RemoteDisconnected):
                    # See https://github.com/psf/requests/issues/4664
                    # The server closed the connection, probably because the Keep Alive has expired
                    # We should be safe to retry, as ClickHouse should not have processed anything on a connection
                    # that it killed.  We also only retry this once, as multiple disconnects are unlikely to be
                    # related to the Keep Alive settings
                    if attempts == 1:
                        logger.debug('Retrying remotely closed connection')
                        attempts = 0
                        continue
                logger.exception('Unexpected Http Driver Exception')
                raise OperationalError(f'Error executing HTTP request {self.url}') from ex
            if 200 <= response.status < 300:
                return response
            if response.status in (429, 503, 504):
                if attempts > retries:
                    self._error_handler(response, True)
                logger.debug('Retrying requests with status code %d', response.status)
            else:
                if error_handler:
                    error_handler(response)
                self._error_handler(response)

    def ping(self):
        """
        See BaseClient doc_string for this method
        """
        try:
            response = self.http.request('GET', f'{self.url}/ping', timeout=3)
            return 200 <= response.status < 300
        except HTTPError:
            logger.debug('ping failed', exc_info=True)
            return False

    def raw_query(self,
                  query: str,
                  parameters: Optional[Union[Sequence, Dict[str, Any]]] = None,
                  settings: Optional[Dict[str, Any]] = None,
                  fmt: str = None) -> bytes:
        """
        See BaseClient doc_string for this method
        """
        final_query, bind_params = bind_query(query, parameters, self.server_tz)
        if fmt:
            final_query += f'\n FORMAT {fmt}'
        params = self._validate_settings(settings or {})
        params.update(bind_params)
        return self._raw_request(final_query, params).data



import re

from typing import Dict, Type, Sequence

from clickhouse_connect.datatypes.base import ClickHouseType, type_map, ch_read_formats, ch_write_formats
from clickhouse_connect.driver.exceptions import ProgrammingError

json_re = re.compile('json', re.IGNORECASE)


def default_encoding(encoding: str):
    ClickHouseType._encoding = encoding  # pylint: disable=protected-access


def set_default_formats(*args, **kwargs):
    fmt_map = format_map(_convert_arguments(*args, **kwargs))
    ch_read_formats.update(fmt_map)
    ch_write_formats.update(fmt_map)


def clear_all_formats():
    ch_read_formats.clear()
    ch_write_formats.clear()


def clear_default_format(pattern: str):
    for ch_type in _matching_types(pattern):
        ch_read_formats.pop(ch_type, None)
        ch_write_formats.pop(ch_type, None)


def set_write_format(pattern: str, fmt: str):
    pattern = json_re.sub('object', pattern)
    for ch_type in _matching_types(pattern):
        ch_write_formats[ch_type] = fmt


def clear_write_format(pattern: str):
    for ch_type in _matching_types(pattern):
        ch_write_formats.pop(ch_type, None)


def set_read_format(pattern: str, fmt: str):
    for ch_type in _matching_types(pattern):
        ch_read_formats[ch_type] = fmt


def clear_read_format(pattern: str):
    for ch_type in _matching_types(pattern):
        ch_read_formats.pop(ch_type, None)


def format_map(fmt_map: Dict[str, str]) -> Dict[Type[ClickHouseType], str]:
    if not fmt_map:
        return {}
    final_map = {}
    for pattern, fmt in fmt_map.items():
        for ch_type in _matching_types(pattern, fmt):
            final_map[ch_type] = fmt
    return final_map


def _convert_arguments(*args, **kwargs) -> Dict[str, str]:
    fmt_map = {}
    try:
        for x in range(0, len(args), 2):
            fmt_map[args[x]] = args[x + 1]
    except (IndexError, TypeError, ValueError) as ex:
        raise ProgrammingError('Invalid type/format arguments for format method') from ex
    fmt_map.update(kwargs)
    return fmt_map


def _matching_types(pattern: str, fmt: str = None) -> Sequence[Type[ClickHouseType]]:
    re_pattern = re.compile(pattern.replace('*', '.*'), re.IGNORECASE)
    matches = [ch_type for type_name, ch_type in type_map.items() if re_pattern.match(type_name)]
    if not matches:
        ProgrammingError(f'Unrecognized ClickHouse type {pattern} when setting formats')
    if fmt:
        invalid = [ch_type.__name__ for ch_type in matches if fmt not in ch_type.valid_formats]
        if invalid:
            raise ProgrammingError(f"{fmt} is not a valid format for ClickHouse types {','.join(invalid)}.")
    return matches



from typing import Optional, Dict, Union, Any

from clickhouse_connect.datatypes.base import ClickHouseType
from clickhouse_connect.datatypes.container import Array
from clickhouse_connect.datatypes.format import format_map
from clickhouse_connect.driver.threads import query_settings


class BaseQueryContext:
    def __init__(self,
                 settings: Optional[Dict[str, Any]] = None,
                 query_formats: Optional[Dict[str, str]] = None,
                 column_formats: Optional[Dict[str, Union[str, Dict[str, str]]]] = None,
                 encoding: Optional[str] = None):
        self.settings = settings or {}
        self.query_formats = query_formats or {}
        self.column_formats = column_formats or {}
        self.encoding = encoding

    def __enter__(self):
        query_settings.query_overrides = format_map(self.query_formats)
        query_settings.query_encoding = self.encoding
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        query_settings.query_overrides = None
        query_settings.column_overrides = None
        query_settings.query_encoding = None

    def start_column(self, name: str, ch_type: ClickHouseType):
        if name in self.column_formats:
            fmts = self.column_formats[name]
            if isinstance(fmts, str):
                if isinstance(ch_type, Array):
                    fmt_map = {ch_type.element_type.__class__: fmts}
                else:
                    fmt_map = {ch_type.__class__: fmts}
            else:
                fmt_map = format_map(fmts)
            query_settings.column_overrides = fmt_map
        else:
            query_settings.column_overrides = None



from abc import ABC, abstractmethod
from typing import Sequence, Any

Matrix = Sequence[Sequence[Any]]


class Closable(ABC):
    @abstractmethod
    def close(self):
        pass


class ByteSource(Closable):

    @abstractmethod
    def read_leb128(self) -> int:
        pass

    @abstractmethod
    def read_leb128_str(self) -> str:
        pass

    @abstractmethod
    def read_uint64(self) -> int:
        pass

    @abstractmethod
    def read_bytes(self, sz: int) -> bytes:
        pass

    @abstractmethod
    def read_str_col(self, num_rows: int, encoding: str):
        pass

    @abstractmethod
    def read_bytes_col(self, sz: int, num_rows: int):
        pass

    @abstractmethod
    def read_fixed_str_col(self, sz: int, num_rows: int, encoding: str):
        pass

    @abstractmethod
    def read_array(self, array_type: str, num_rows: int):
        pass

    @abstractmethod
    def read_byte(self) -> int:
        pass



from typing import Union, Sequence, MutableSequence
from uuid import UUID as PYUUID, SafeUUID

from clickhouse_connect.datatypes.base import TypeDef, ClickHouseType, ArrayType, UnsupportedType
from clickhouse_connect.datatypes.registry import get_from_name
from clickhouse_connect.driver.types import ByteSource

empty_uuid_b = bytes(b'\x00' * 16)


class UUID(ClickHouseType):
    valid_formats = 'string', 'native'

    @property
    def python_null(self):
        return '' if self.read_format() == 'string' else PYUUID(int=0)

    def np_type(self, _str_len: int = 0):
        return 'U36' if self.read_format() == 'string' else 'O'

    def _read_native_binary(self, source: ByteSource, num_rows: int):
        if self.read_format() == 'string':
            return self._read_native_str(source, num_rows)
        return self._read_native_uuid(source, num_rows)

    # pylint: disable=too-many-locals
    @staticmethod
    def _read_native_uuid(source: ByteSource, num_rows: int):
        v = source.read_array('Q', num_rows * 2)
        empty_uuid = PYUUID(int=0)
        new_uuid = PYUUID.__new__
        unsafe = SafeUUID.unsafe
        oset = object.__setattr__
        column = []
        app = column.append
        for i in range(num_rows):
            ix = i << 1
            int_value = v[ix] << 64 | v[ix + 1]
            if int_value == 0:
                app(empty_uuid)
            else:
                fast_uuid = new_uuid(PYUUID)
                oset(fast_uuid, 'int', int_value)
                oset(fast_uuid, 'is_safe', unsafe)
                app(fast_uuid)
        return column

    @staticmethod
    def _read_native_str(source: ByteSource, num_rows: int):
        v = source.read_array('Q', num_rows * 2)
        column = []
        app = column.append
        for i in range(num_rows):
            ix = i << 1
            x = f'{(v[ix] << 64 | v[ix + 1]):032x}'
            app(f'{x[:8]}-{x[8:12]}-{x[12:16]}-{x[16:20]}-{x[20:]}')
        return column

    # pylint: disable=too-many-branches
    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        first = self._first_value(column)
        empty = empty_uuid_b
        if isinstance(first, str) or self.write_format() == 'string':
            for v in column:
                if v:
                    x = int(v, 16)
                    dest += (x >> 64).to_bytes(8, 'little') + (x & 0xffffffffffffffff).to_bytes(8, 'little')
                else:
                    dest += empty
        elif isinstance(first, int):
            for x in column:
                if x:
                    dest += (x >> 64).to_bytes(8, 'little') + (x & 0xffffffffffffffff).to_bytes(8, 'little')
                else:
                    dest += empty
        elif isinstance(first, PYUUID):
            for v in column:
                if v:
                    x = v.int
                    dest += (x >> 64).to_bytes(8, 'little') + (x & 0xffffffffffffffff).to_bytes(8, 'little')
                else:
                    dest += empty
        elif isinstance(first, (bytes, bytearray, memoryview)):
            for v in column:
                if v:
                    dest += bytes(reversed(v[:8])) + bytes(reversed(v[8:]))
                else:
                    dest += empty
        else:
            dest += empty * len(column)


class Nothing(ArrayType):
    _array_type = 'b'

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self.nullable = True

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        dest += bytes(0x30 for _ in range(len(column)))


class SimpleAggregateFunction(ClickHouseType):
    _slots = ('element_type',)

    def __init__(self, type_def: TypeDef):
        super().__init__(type_def)
        self.element_type: ClickHouseType = get_from_name(type_def.values[1])
        self._name_suffix = type_def.arg_str

    def _read_native_binary(self, source: ByteSource, num_rows: int):
        return self.element_type.read_native_data(source, num_rows)

    def _write_native_binary(self, column: Union[Sequence, MutableSequence], dest: MutableSequence):
        self.element_type.write_native_data(column, dest)


class AggregateFunction(UnsupportedType):
    pass
